@@ -68,6 +68,7 @@ struct symbol
 {
 	struct symbol *next;
 	struct module *owner;
+	uint64_t ver;
 	char name[0];
 };
 
@@ -86,12 +87,13 @@ static inline unsigned int tdb_hash(const char *name)
 	return (1103515243 * value + 12345);
 }
 
-void add_symbol(const char *name, struct module *owner)
+void add_symbol(const char *name, uint64_t ver, struct module *owner)
 {
 	unsigned int hash;
 	struct symbol *new = NOFAIL(malloc(sizeof *new + strlen(name) + 1));
 
 	new->owner = owner;
+	new->ver = ver;
 	strcpy(new->name, name);
 
 	hash = tdb_hash(name) % SYMBOL_HASH_SIZE;
@@ -99,9 +101,10 @@ void add_symbol(const char *name, struct module *owner)
 	symbolhash[hash] = new;
 }
 
-static int print_unknown;
+static int print_unknown, check_symvers;
 
-struct module *find_symbol(const char *name, const char *modname, int weak)
+struct module *find_symbol(const char *name, uint64_t ver,
+		const char *modname, int weak)
 {
 	struct symbol *s;
 
@@ -111,7 +114,13 @@ struct module *find_symbol(const char *name, const char *modname, int weak)
 
 	for (s = symbolhash[tdb_hash(name) % SYMBOL_HASH_SIZE]; s; s=s->next) {
 		if (streq(s->name, name))
-			return s->owner;
+			break;
+	}
+	if (s) {
+		if (ver && s->ver && s->ver != ver && print_unknown && !weak)
+			warn("%s disagrees about version of symbol %s\n",
+					modname, name);
+		return s->owner;
 	}
 
 	if (print_unknown && !weak)
@@ -132,6 +141,14 @@ void add_dep(struct module *mod, struct module *depends_on)
 	mod->deps[mod->num_deps++] = depends_on;
 }
 
+static void add_fake_syms(void)
+{
+	/* __this_module is magic inserted by kernel loader. */
+	add_symbol("__this_module", 0, NULL);
+	/* On S390, this is faked up too */
+	add_symbol("_GLOBAL_OFFSET_TABLE_", 0, NULL);
+}
+
 static void load_system_map(const char *filename)
 {
 	FILE *system_map;
@@ -158,32 +175,56 @@ static void load_system_map(const char *filename)
 
 		/* Covers gpl-only and normal symbols. */
 		if (strstarts(ptr+1, ksymstr))
-			add_symbol(ptr+1+ksymstr_len, NULL);
+			add_symbol(ptr+1+ksymstr_len, 0, NULL);
 	}
 
 	fclose(system_map);
+	add_fake_syms();
+}
 
-	/* __this_module is magic inserted by kernel loader. */
-	add_symbol("__this_module", NULL);
-	/* On S390, this is faked up too */
-	add_symbol("_GLOBAL_OFFSET_TABLE_", NULL);
+static void load_module_symvers(const char *filename)
+{
+	FILE *module_symvers;
+	char line[10240];
+
+	module_symvers = fopen(filename, "r");
+	if (!module_symvers)
+		fatal("Could not open '%s': %s\n", filename, strerror(errno));
+
+	/* eg. "0xb352177e\tfind_first_bit\tvmlinux\tEXPORT_SYMBOL" */
+	while (fgets(line, sizeof(line)-1, module_symvers)) {
+		const char *ver, *sym, *where;
+
+		ver = strtok(line, " \t");
+		sym = strtok(NULL, " \t");
+		where = strtok(NULL, " \t");
+		if (!ver || !sym || !where)
+			continue;
+
+		if (streq(where, "vmlinux"))
+			add_symbol(sym, strtoull(ver, NULL, 16), NULL);
+	}
+
+	fclose(module_symvers);
+	add_fake_syms();
 }
 
 static struct option options[] = { { "all", 0, NULL, 'a' },
 				   { "quick", 0, NULL, 'A' },
 				   { "basedir", 1, NULL, 'b' },
-				   { "errsyms", 0, NULL, 'e' },
+				   { "config", 1, NULL, 'C' },
+				   { "symvers", 1, NULL, 'E' },
 				   { "filesyms", 1, NULL, 'F' },
-				   { "help", 0, NULL, 'h' },
-				   { "show", 0, NULL, 'n' },
-				   { "dry-run", 0, NULL, 'n' },
+				   { "errsyms", 0, NULL, 'e' },
+				   { "unresolved-error", 0, NULL, 'u' },
 				   { "quiet", 0, NULL, 'q' },
 				   { "root", 0, NULL, 'r' },
-				   { "unresolved-error", 0, NULL, 'u' },
 				   { "verbose", 0, NULL, 'v' },
+				   { "show", 0, NULL, 'n' },
+				   { "dry-run", 0, NULL, 'n' },
+				   { "help", 0, NULL, 'h' },
 				   { "version", 0, NULL, 'V' },
-				   { "config", 1, NULL, 'C' },
-				   { "warn", 1, NULL, 'w' },
+				   { "warn", 0, NULL, 'w' },
 				   { "map", 0, NULL, 'm' },
 				   { NULL, 0, NULL, 0 } };
 
@@ -242,7 +283,10 @@ static void print_usage(const char *name)
 	"\t    --basedir basedirectory    Use an image of a module tree.\n"
 	"\t-F kernelsyms\n"
 	"\t    --filesyms kernelsyms      Use the file instead of the\n"
-	"\t                               current kernel symbols.\n",
+	"\t                               current kernel symbols.\n"
+	"\t-E Module.symvers\n"
+	"\t    --symvers Module.symvers   Use Module.symvers file to check\n"
+	"\t                               symbol versions.\n",
 	"depmod", "depmod");
 }
 
@@ -390,7 +434,7 @@ static const char *compress_path(const char *path, const char *basedir)
 	return path;
 }
 
-static void output_deps(struct module *modules,
+static int output_deps(struct module *modules,
 			FILE *out, char *dirname)
 {
 	struct module *i;
@@ -409,12 +453,13 @@ static void output_deps(struct module *modules,
 		}
 		fprintf(out, "\n");
 	}
+	return 1;
 }
 
 /* warn whenever duplicate module aliases, deps, or symbols are found. */
 int warn_dups = 0;
 
-static void output_deps_bin(struct module *modules,
+static int output_deps_bin(struct module *modules,
 			FILE *out, char *dirname)
 {
 	struct module *i;
@@ -451,6 +496,8 @@ static void output_deps_bin(struct module *modules,
 	
 	index_write(index, out);
 	index_destroy(index);
+
+	return 1;
 }
 
 
@@ -650,24 +697,28 @@ static void calculate_deps(struct module *module)
 	unsigned int i;
 	struct string_table *symnames;
 	struct string_table *symtypes;
+	uint64_t *symvers = NULL;
 	struct elf_file *file;
 
 	module->num_deps = 0;
 	module->deps = NULL;
 	file = module->file;
 
-	symnames = file->ops->load_dep_syms(module->pathname, file, &symtypes);
+	symnames = file->ops->load_dep_syms(file, &symtypes,
+			check_symvers ? &symvers : NULL);
 	if (!symnames || !symtypes)
 		return;
 
 	for (i = 0; i < symnames->cnt; i++) {
 		const char *name;
+		uint64_t ver;
 		struct module *owner;
 		int weak;
 
 		name = symnames->str[i];
+		ver = symvers ? symvers[i] : 0;
 		weak = (*(symtypes->str[i]) == 'W');
-		owner = find_symbol(name, module->pathname, weak);
+		owner = find_symbol(name, ver, module->pathname, weak);
 		if (owner) {
 			info("%s needs \"%s\": %s\n",
 			       module->pathname, name,
@@ -678,6 +729,7 @@ static void calculate_deps(struct module *module)
 
 	free(symnames);
 	free(symtypes);
+	free(symvers);
 }
 
 static struct module *parse_modules(struct module *list)
@@ -688,13 +740,17 @@ static struct module *parse_modules(struct module *list)
 	int j;
 
 	for (i = list; i; i = i->next) {
+		uint64_t *symvers = NULL;
 		file = i->file;
-		syms = file->ops->load_symbols(file);
+		syms = file->ops->load_symbols(file,
+				check_symvers ? &symvers : NULL);
 		if (syms) {
 			for (j = 0; j < syms->cnt; j++)
-				add_symbol(syms->str[j], i);
+				add_symbol(syms->str[j],
+					symvers ? symvers[j] : 0, i);
 			strtbl_free(syms);
 		}
+		free(symvers);
 		file->ops->fetch_tables(file, &i->tables);
 	}
 	
@@ -716,7 +772,7 @@ static struct module *parse_modules(struct module *list)
 }
 
 /* Simply dump hash table. */
-static void output_symbols(struct module *unused, FILE *out, char *dirname)
+static int output_symbols(struct module *unused, FILE *out, char *dirname)
 {
 	unsigned int i;
 
@@ -733,9 +789,10 @@ static void output_symbols(struct module *unused, FILE *out, char *dirname)
 			}
 		}
 	}
+	return 1;
 }
 
-static void output_symbols_bin(struct module *unused, FILE *out, char *dirname)
+static int output_symbols_bin(struct module *unused, FILE *out, char *dirname)
 {
 	struct index_node *index;
 	unsigned int i;
@@ -764,9 +821,47 @@ static void output_symbols_bin(struct module *unused, FILE *out, char *dirname)
 	
 	index_write(index, out);
 	index_destroy(index);
+
+	return 1;
 }
 
-static void output_aliases(struct module *modules, FILE *out, char *dirname)
+static int output_builtin_bin(struct module *unused, FILE *out, char *dirname)
+{
+	struct index_node *index;
+	char *textfile, *line;
+	unsigned int linenum;
+	FILE *f;
+
+	nofail_asprintf(&textfile, "%s/modules.builtin", dirname);
+	if (!(f = fopen(textfile, "r"))) {
+		if (errno != ENOENT)
+			fatal("Could not open '%s': %s\n",
+					textfile, strerror(errno));
+		free(textfile);
+		return 0;
+	}
+	free(textfile);
+	index = index_create();
+
+	while ((line = getline_wrapped(f, &linenum)) != NULL) {
+		char *module = line;
+
+		if (!*line || *line == '#') {
+			free(line);
+			continue;
+		}
+		filename2modname(module, module);
+		index_insert(index, module, "", 0);
+		free(line);
+	}
+	fclose(f);
+	index_write(index, out);
+	index_destroy(index);
+
+	return 1;
+}
+
+static int output_aliases(struct module *modules, FILE *out, char *dirname)
 {
 	struct module *i;
 	struct elf_file *file;
@@ -796,9 +891,10 @@ static void output_aliases(struct module *modules, FILE *out, char *dirname)
 		}
 		strtbl_free(tbl);
 	}
+	return 1;
 }
 
-static void output_aliases_bin(struct module *modules, FILE *out, char *dirname)
+static int output_aliases_bin(struct module *modules, FILE *out, char *dirname)
 {
 	struct module *i;
 	struct elf_file *file;
@@ -848,11 +944,13 @@ static void output_aliases_bin(struct module *modules, FILE *out, char *dirname)
 	
 	index_write(index, out);
 	index_destroy(index);
+
+	return 1;
 }
 
 struct depfile {
 	char *name;
-	void (*func)(struct module *, FILE *, char *dirname);
+	int (*func)(struct module *, FILE *, char *dirname);
 	int map_file;
 };
 
@@ -870,7 +968,8 @@ static struct depfile depfiles[] = {
 	{ "modules.alias", output_aliases, 0 },
 	{ "modules.alias.bin", output_aliases_bin, 0 },
 	{ "modules.symbols", output_symbols, 0 },
-	{ "modules.symbols.bin", output_symbols_bin, 0 }
+	{ "modules.symbols.bin", output_symbols_bin, 0 },
+	{ "modules.builtin.bin", output_builtin_bin, 0 },
 };
 
 /* If we can't figure it out, it's safe to say "true". */
@@ -1175,29 +1274,33 @@ struct module_overrides *overrides = NULL;
 int main(int argc, char *argv[])
 {
 	int opt, all = 0, maybe_all = 0, doing_stdout = 0;
-	char *basedir = "", *dirname, *version, *badopt = NULL,
-		*system_map = NULL;
+	char *basedir = "", *dirname, *version;
+	char *system_map = NULL, *module_symvers = NULL;
 	int i;
 	const char *config = NULL;
 
 	if (native_endianness() == 0)
 		abort();
 
-	/* Don't print out any errors just yet, we might want to exec
-           backwards compat version. */
-	opterr = 0;
-	while ((opt = getopt_long(argc, argv, "ab:ArehnqruvVF:C:wm", options, NULL))
+	while ((opt = getopt_long(argc, argv, "aAb:C:E:F:euqrvnhVwm", options, NULL))
 	       != -1) {
 		switch (opt) {
 		case 'a':
 			all = 1;
 			break;
+		case 'A':
+			maybe_all = 1;
+			break;
 		case 'b':
 			basedir = optarg;
 			skipchars = strlen(basedir);
 			break;
-		case 'A':
-			maybe_all = 1;
+		case 'C':
+			config = optarg;
+			break;
+		case 'E':
+			module_symvers = optarg;
+			check_symvers = 1;
 			break;
 		case 'F':
 			system_map = optarg;
@@ -1205,23 +1308,20 @@ int main(int argc, char *argv[])
 		case 'e':
 			print_unknown = 1;
 			break;
-		case 'v':
-			verbose = 1;
-			break;
 		case 'u':
 		case 'q':
 		case 'r':
 			break;
-		case 'C':
-			config = optarg;
+		case 'v':
+			verbose = 1;
+			break;
+		case 'n':
+			doing_stdout = 1;
 			break;
 		case 'h':
 			print_usage(argv[0]);
 			exit(0);
 			break;
-		case 'n':
-			doing_stdout = 1;
-			break;
 		case 'V':
 			printf("%s %s\n", PACKAGE, VERSION);
 			exit(0);
@@ -1232,15 +1332,19 @@ int main(int argc, char *argv[])
 			force_map_files = 1;
 			break;
 		default:
-			badopt = argv[optind-1];
+			print_usage(argv[0]);
+			exit(1);
 		}
 	}
 
-	/* We can't print unknowns without a System.map */
-	if (!system_map)
-		print_unknown = 0;
-	else
+	if (module_symvers)
+		load_module_symvers(module_symvers);
+	else if (system_map)
 		load_system_map(system_map);
+	else if (print_unknown) {
+		warn("-e needs -E or -F");
+		print_unknown = 0;
+	}
 
 	/* They can specify the version naked on the command line */
 	if (optind < argc && is_version_number(argv[optind])) {
@@ -1259,13 +1363,6 @@ int main(int argc, char *argv[])
 		exit(2);
 	}
 
-	if (badopt) {
-		fprintf(stderr, "%s: malformed/unrecognized option '%s'\n",
-			argv[0], badopt);
-		print_usage(argv[0]);
-		exit(1);
-	}
-
 	/* Depmod -a by default if no names. */
 	if (optind == argc)
 		all = 1;
@@ -1317,6 +1414,7 @@ int main(int argc, char *argv[])
 
 	for (i = 0; i < sizeof(depfiles)/sizeof(depfiles[0]); i++) {
 		FILE *out;
+		int res;
 		struct depfile *d = &depfiles[i];
 		char depname[strlen(dirname) + 1 + strlen(d->name) + 1];
 		char tmpname[strlen(dirname) + 1 + strlen(d->name) +
@@ -1337,12 +1435,18 @@ int main(int argc, char *argv[])
 			if (ends_in(depname, ".bin"))
 				continue;
 		}
-		d->func(list, out, dirname);
-		if (!doing_stdout) {
-			fclose(out);
+		res = d->func(list, out, dirname);
+		if (doing_stdout)
+			continue;
+		fclose(out);
+		if (res) {
 			if (rename(tmpname, depname) < 0)
 				fatal("Could not rename %s into %s: %s\n",
 					tmpname, depname, strerror(errno));
+		} else {
+			if (unlink(tmpname) < 0)
+				warn("Could not delete %s: %s\n",
+					tmpname, strerror(errno));
 		}
 	}
 
