@@ -40,6 +40,7 @@
 #include <syslog.h>
 
 #include "util.h"
+#include "elfops.h"
 #include "zlibsupport.h"
 #include "logging.h"
 #include "index.h"
@@ -59,12 +60,24 @@ struct module {
 	char filename[0];
 };
 
+typedef enum
+{
+	mit_remove = 1,
+	mit_dry_run = 2,
+	mit_first_time = 4,
+	mit_use_blacklist = 8,
+	mit_ignore_commands = 16,
+	mit_ignore_loaded = 32,
+	mit_strip_vermagic = 64,
+	mit_strip_modversion = 128,
+	mit_resolve_alias = 256
+
+} modprobe_flags_t;
+
 #ifndef MODULE_DIR
 #define MODULE_DIR "/lib/modules"
 #endif
 
-typedef void (*errfn_t)(const char *fmt, ...);
-
 static void print_usage(const char *progname)
 {
 	fprintf(stderr,
@@ -160,10 +173,7 @@ static int add_modules_dep_line(char *line,
 
 	/* Is this the module we are looking for? */
 	*ptr = '\0';
-	if (strrchr(line, '/'))
-		modname = strrchr(line, '/') + 1;
-	else
-		modname = line;
+	modname = my_basename(line);
 
 	len = strlen(modname);
 	if (strchr(modname, '.'))
@@ -285,7 +295,7 @@ static const char *remove_moderror(int err)
 	}
 }
 
-static void replace_modname(struct module *module,
+static void replace_modname(struct elf_file *module,
 			    void *mem, unsigned long len,
 			    const char *oldname, const char *newname)
 {
@@ -303,165 +313,41 @@ static void replace_modname(struct module *module,
 		}
 	}
 
-	warn("Could not find old name in %s to replace!\n", module->filename);
+	warn("Could not find old name in %s to replace!\n", module->pathname);
 }
 
-static void *get_section32(void *file,
-			   unsigned long size,
-			   const char *name,
-			   unsigned long *secsize)
-{
-	Elf32_Ehdr *hdr = file;
-	Elf32_Shdr *sechdrs = file + hdr->e_shoff;
-	const char *secnames;
-	unsigned int i;
-
-	/* Too short? */
-	if (size < sizeof(*hdr))
-		return NULL;
-	if (size < hdr->e_shoff + hdr->e_shnum * sizeof(sechdrs[0]))
-		return NULL;
-	if (size < sechdrs[hdr->e_shstrndx].sh_offset)
-		return NULL;
-		
-	secnames = file + sechdrs[hdr->e_shstrndx].sh_offset;
-	for (i = 1; i < hdr->e_shnum; i++)
-		if (streq(secnames + sechdrs[i].sh_name, name)) {
-			*secsize = sechdrs[i].sh_size;
-			return file + sechdrs[i].sh_offset;
-		}
-	return NULL;
-}
-
-static void *get_section64(void *file,
-			   unsigned long size,
-			   const char *name,
-			   unsigned long *secsize)
-{
-	Elf64_Ehdr *hdr = file;
-	Elf64_Shdr *sechdrs = file + hdr->e_shoff;
-	const char *secnames;
-	unsigned int i;
-
-	/* Too short? */
-	if (size < sizeof(*hdr))
-		return NULL;
-	if (size < hdr->e_shoff + hdr->e_shnum * sizeof(sechdrs[0]))
-		return NULL;
-	if (size < sechdrs[hdr->e_shstrndx].sh_offset)
-		return NULL;
-		
-	secnames = file + sechdrs[hdr->e_shstrndx].sh_offset;
-	for (i = 1; i < hdr->e_shnum; i++)
-		if (streq(secnames + sechdrs[i].sh_name, name)) {
-			*secsize = sechdrs[i].sh_size;
-			return file + sechdrs[i].sh_offset;
-		}
-	return NULL;
-}
-
-static int elf_ident(void *mod, unsigned long size)
-{
-	/* "\177ELF" <byte> where byte = 001 for 32-bit, 002 for 64 */
-	char *ident = mod;
-
-	if (size < EI_CLASS || memcmp(mod, ELFMAG, SELFMAG) != 0)
-		return ELFCLASSNONE;
-	return ident[EI_CLASS];
-}
-
-static void *get_section(void *file,
-			 unsigned long size,
-			 const char *name,
-			 unsigned long *secsize)
-{
-	switch (elf_ident(file, size)) {
-	case ELFCLASS32:
-		return get_section32(file, size, name, secsize);
-	case ELFCLASS64:
-		return get_section64(file, size, name, secsize);
-	default:
-		return NULL;
-	}
-}
-
-static void rename_module(struct module *module,
-			  void *mod,
-			  unsigned long len,
+static void rename_module(struct elf_file *module,
+			  const char *oldname,
 			  const char *newname)
 {
 	void *modstruct;
-	unsigned long modstruct_len;
+	unsigned long len;
 
 	/* Old-style */
-	modstruct = get_section(mod, len, ".gnu.linkonce.this_module",
-				&modstruct_len);
+	modstruct = module->ops->load_section(module,
+		".gnu.linkonce.this_module", &len);
 	/* New-style */
 	if (!modstruct)
-		modstruct = get_section(mod, len, "__module", &modstruct_len);
+		modstruct = module->ops->load_section(module, "__module", &len);
 	if (!modstruct)
 		warn("Could not find module name to change in %s\n",
-		     module->filename);
+		     module->pathname);
 	else
-		replace_modname(module, modstruct, modstruct_len,
-				module->modname, newname);
+		replace_modname(module, modstruct, len, oldname, newname);
 }
 
-/* Kernel told to ignore these sections if SHF_ALLOC not set. */
-static void invalidate_section32(void *mod, const char *secname)
+static void clear_magic(struct elf_file *module)
 {
-	Elf32_Ehdr *hdr = mod;
-	Elf32_Shdr *sechdrs = mod + hdr->e_shoff;
-	const char *secnames = mod + sechdrs[hdr->e_shstrndx].sh_offset;
-	unsigned int i;
-
-	for (i = 1; i < hdr->e_shnum; i++)
-		if (streq(secnames+sechdrs[i].sh_name, secname))
-			sechdrs[i].sh_flags &= ~SHF_ALLOC;
-}
-
-static void invalidate_section64(void *mod, const char *secname)
-{
-	Elf64_Ehdr *hdr = mod;
-	Elf64_Shdr *sechdrs = mod + hdr->e_shoff;
-	const char *secnames = mod + sechdrs[hdr->e_shstrndx].sh_offset;
-	unsigned int i;
-
-	for (i = 1; i < hdr->e_shnum; i++)
-		if (streq(secnames+sechdrs[i].sh_name, secname))
-			sechdrs[i].sh_flags &= ~(unsigned long long)SHF_ALLOC;
-}
-
-static void strip_section(struct module *module,
-			  void *mod,
-			  unsigned long len,
-			  const char *secname)
-{
-	switch (elf_ident(mod, len)) {
-	case ELFCLASS32:
-		invalidate_section32(mod, secname);
-		break;
-	case ELFCLASS64:
-		invalidate_section64(mod, secname);
-		break;
-	default:
-		warn("Unknown module format in %s: not forcing version\n",
-		     module->filename);
-	}
-}
-
-static void clear_magic(struct module *module, void *mod, unsigned long len)
-{
-	const char *p;
-	unsigned long modlen;
+	struct string_table *tbl;
+	int j;
 
 	/* Old-style: __vermagic section */
-	strip_section(module, mod, len, "__vermagic");
+	module->ops->strip_section(module, "__vermagic");
 
 	/* New-style: in .modinfo section */
-	for (p = get_section(mod, len, ".modinfo", &modlen);
-	     p;
-	     p = next_string(p, &modlen)) {
+	tbl = module->ops->load_strings(module, ".modinfo", NULL, fatal);
+	for (j = 0; tbl && j < tbl->cnt; j++) {
+		const char *p = tbl->str[j];
 		if (strstarts(p, "vermagic=")) {
 			memset((char *)p, 0, strlen(p));
 			return;
@@ -652,6 +538,24 @@ static int read_attribute(const char *filename, char *buf, size_t buflen)
 	return (s == NULL) ? -1 : 1;
 }
 
+/* is this a built-in module?
+ * 0: no, 1: yes, -1: don't know
+ */
+static int module_builtin(const char *dirname, const char *modname)
+{
+	struct index_file *index;
+	char *filename, *value;
+
+	nofail_asprintf(&filename, "%s/modules.builtin.bin", dirname);
+	index = index_file_open(filename);
+	free(filename);
+	if (!index)
+		return -1;
+	value = index_search(index, modname);
+	free(value);
+	return value ? 1 : 0;
+}
+
 /* Is module in /sys/module?  If so, fill in usecount if not NULL. 
    0 means no, 1 means yes, -1 means unknown.
  */
@@ -701,273 +605,20 @@ static int module_in_kernel(const char *modname, unsigned int *usecount)
 	return 1;
 }
 
-/* Do an install/remove command: replace $CMDLINE_OPTS if it's specified. */
-static void do_command(const char *modname,
-		       const char *command,
-		       int dry_run,
-		       errfn_t error,
-		       const char *type,
-		       const char *cmdline_opts)
-{
-	int ret;
-	char *p, *replaced_cmd = NOFAIL(strdup(command));
-
-	while ((p = strstr(replaced_cmd, "$CMDLINE_OPTS")) != NULL) {
-		char *new;
-		nofail_asprintf(&new, "%.*s%s%s",
-			 (int)(p - replaced_cmd), replaced_cmd, cmdline_opts,
-			 p + strlen("$CMDLINE_OPTS"));
-		free(replaced_cmd);
-		replaced_cmd = new;
-	}
-
-	info("%s %s\n", type, replaced_cmd);
-	if (dry_run)
-		return;
-
-	setenv("MODPROBE_MODULE", modname, 1);
-	ret = system(replaced_cmd);
-	if (ret == -1 || WEXITSTATUS(ret))
-		error("Error running %s command for %s\n", type, modname);
-	free(replaced_cmd);
-}
-
-/* Actually do the insert.  Frees second arg. */
-static int insmod(struct list_head *list,
-		   char *optstring,
-		   const char *newname,
-		   int first_time,
-		   errfn_t error,
-		   int dry_run,
-		   const struct module_options *options,
-		   const struct module_command *commands,
-		   int ignore_commands,
-		   int ignore_proc,
-		   int strip_vermagic,
-		   int strip_modversion,
-		   const char *cmdline_opts)
-{
-	int ret, fd;
-	unsigned long len;
-	void *map;
-	const char *command;
-	struct module *mod = list_entry(list->next, struct module, list);
-	int rc = 0;
-
-	/* Take us off the list. */
-	list_del(&mod->list);
-
-	/* Do things we (or parent) depend on first. */
-	if (!list_empty(list)) {
-		if ((rc = insmod(list, NOFAIL(strdup("")), NULL, 0, warn,
-		       dry_run, options, commands, 0, ignore_proc,
-		       strip_vermagic, strip_modversion, "")) != 0) {
-			error("Error inserting %s (%s): %s\n",
-				mod->modname, mod->filename,
-				insert_moderror(errno));
-			goto out_optstring;
-		}
-	}
-
-	fd = open_file(mod->filename);
-	if (fd < 0) {
-		error("Could not open '%s': %s\n",
-		      mod->filename, strerror(errno));
-		goto out_optstring;
-	}
-
-	/* Don't do ANYTHING if already in kernel. */
-	if (!ignore_proc
-	    && module_in_kernel(newname ?: mod->modname, NULL) == 1) {
-		if (first_time)
-			error("Module %s already in kernel.\n",
-			      newname ?: mod->modname);
-		goto out_unlock;
-	}
-
-	command = find_command(mod->modname, commands);
-	if (command && !ignore_commands) {
-		close_file(fd);
-		do_command(mod->modname, command, dry_run, error,
-			   "install", cmdline_opts);
-		goto out_optstring;
-	}
-
-	map = grab_fd(fd, &len);
-	if (!map) {
-		error("Could not read '%s': %s\n",
-		      mod->filename, strerror(errno));
-		goto out_unlock;
-	}
-
-	/* Rename it? */
-	if (newname)
-		rename_module(mod, map, len, newname);
-
-	if (strip_modversion)
-		strip_section(mod, map, len, "__versions");
-	if (strip_vermagic)
-		clear_magic(mod, map, len);
-
-	/* Config file might have given more options */
-	optstring = add_extra_options(mod->modname, optstring, options);
-
-	info("insmod %s %s\n", mod->filename, optstring);
-
-	if (dry_run)
-		goto out;
-
-	ret = init_module(map, len, optstring);
-	if (ret != 0) {
-		if (errno == EEXIST) {
-			if (first_time)
-				error("Module %s already in kernel.\n",
-				      newname ?: mod->modname);
-			goto out_unlock;
-		}
-		/* don't warn noisely if we're loading multiple aliases. */
-		/* one of the aliases may try to use hardware we don't have. */
-		if ((error != warn) || (verbose))
-			error("Error inserting %s (%s): %s\n",
-			      mod->modname, mod->filename,
-			      insert_moderror(errno));
-			rc = 1;
-	}
- out:
-	release_file(map, len);
- out_unlock:
-	close_file(fd);
- out_optstring:
-	free(optstring);
-	return rc;
-}
-
-/* Do recursive removal. */
-static void rmmod(struct list_head *list,
-		  const char *name,
-		  int first_time,
-		  errfn_t error,
-		  int dry_run,
-		  struct module_command *commands,
-		  int ignore_commands,
-		  int ignore_inuse,
-		  const char *cmdline_opts,
-		  int flags)
-{
-	const char *command;
-	unsigned int usecount = 0;
-	struct module *mod = list_entry(list->next, struct module, list);
-
-	/* Take first one off the list. */
-	list_del(&mod->list);
-
-	if (!name)
-		name = mod->modname;
-
-	/* Even if renamed, find commands to orig. name. */
-	command = find_command(mod->modname, commands);
-	if (command && !ignore_commands) {
-		do_command(mod->modname, command, dry_run, error,
-			   "remove", cmdline_opts);
-		goto remove_rest;
-	}
-
-	if (module_in_kernel(name, &usecount) == 0)
-		goto nonexistent_module;
-
-	if (usecount != 0) {
-		if (!ignore_inuse)
-			error("Module %s is in use.\n", name);
-		goto remove_rest;
-	}
-
-	info("rmmod %s\n", mod->filename);
-
-	if (dry_run)
-		goto remove_rest;
-
-	if (delete_module(name, O_EXCL) != 0) {
-		if (errno == ENOENT)
-			goto nonexistent_module;
-		error("Error removing %s (%s): %s\n",
-		      name, mod->filename,
-		      remove_moderror(errno));
-	}
-
- remove_rest:
-	/* Now do things we depend. */
-	if (!list_empty(list))
-		rmmod(list, NULL, 0, warn, dry_run, commands,
-		      0, 1, "", flags);
-	return;
-
-nonexistent_module:
-	if (first_time)
-		fatal("Module %s is not in kernel.\n", mod->modname);
-	goto remove_rest;
-}
-
-struct modver32_info
-{
-       uint32_t crc;
-       char name[64 - sizeof(uint32_t)];
-};
-
-struct modver64_info
-{
-       uint64_t crc;
-       char name[64 - sizeof(uint64_t)];
-};
-
-const char *skip_dot(const char *str)
-{
-       /* For our purposes, .foo matches foo.  PPC64 needs this. */
-       if (str && str[0] == '.')
-               return str + 1;
-       return str;
-}
-
 void dump_modversions(const char *filename, errfn_t error)
 {
-       unsigned long size, secsize;
-       void *file = grab_file(filename, &size);
-       struct modver32_info *info32;
-       struct modver64_info *info64;
-       int n;
+	struct elf_file *module;
 
-       if (!file) {
-               error("%s: %s\n", filename, strerror(errno));
-               return;
-       }
-       switch (elf_ident(file, size)) {
-       case ELFCLASS32:
-               info32 = get_section32(file, size, "__versions", &secsize);
-               if (!info32)
-                       return;  /* Does not seem to be a kernel module */
-               if (secsize % sizeof(struct modver32_info))
-                       error("Wrong section size in %s\n", filename);
-               for (n = 0; n < secsize / sizeof(struct modver32_info); n++)
-                       printf("0x%08lx\t%s\n", (unsigned long)
-                              info32[n].crc, skip_dot(info32[n].name));
-               break;
-
-       case ELFCLASS64:
-               info64 = get_section64(file, size, "__versions", &secsize);
-               if (!info64)
-                       return;  /* Does not seem to be a kernel module */
-               if (secsize % sizeof(struct modver64_info))
-                       error("Wrong section size in %s\n", filename);
-               for (n = 0; n < secsize / sizeof(struct modver64_info); n++)
-                       printf("0x%08llx\t%s\n", (unsigned long long)
-                              info64[n].crc, skip_dot(info64[n].name));
-               break;
-
-       default:
-               error("%s: ELF class not recognized\n", filename);
-       }
+	module = grab_elf_file(filename);
+	if (!module) {
+		error("%s: %s\n", filename, strerror(errno));
+		return;
+	}
+	if (module->ops->dump_modvers(module) < 0)
+		error("Wrong section size in '%s'\n", filename);
+	release_elf_file(module);
 }
 
-
 /* Does path contain directory(s) subpath? */
 static int type_matches(const char *path, const char *subpath)
 {
@@ -1417,22 +1068,215 @@ static char *gather_options(char *argv[])
 	return optstring;
 }
 
+/* Do an install/remove command: replace $CMDLINE_OPTS if it's specified. */
+static void do_command(const char *modname,
+		       const char *command,
+		       int dry_run,
+		       errfn_t error,
+		       const char *type,
+		       const char *cmdline_opts)
+{
+	int ret;
+	char *p, *replaced_cmd = NOFAIL(strdup(command));
+
+	while ((p = strstr(replaced_cmd, "$CMDLINE_OPTS")) != NULL) {
+		char *new;
+		nofail_asprintf(&new, "%.*s%s%s",
+			 (int)(p - replaced_cmd), replaced_cmd, cmdline_opts,
+			 p + strlen("$CMDLINE_OPTS"));
+		free(replaced_cmd);
+		replaced_cmd = new;
+	}
+
+	info("%s %s\n", type, replaced_cmd);
+	if (dry_run)
+		return;
+
+	setenv("MODPROBE_MODULE", modname, 1);
+	ret = system(replaced_cmd);
+	if (ret == -1 || WEXITSTATUS(ret))
+		error("Error running %s command for %s\n", type, modname);
+	free(replaced_cmd);
+}
+
+/* Actually do the insert.  Frees second arg. */
+static int insmod(struct list_head *list,
+		   char *optstring,
+		   const char *newname,
+		   const struct module_options *options,
+		   const struct module_command *commands,
+		   const char *cmdline_opts,
+		   errfn_t error,
+		   modprobe_flags_t flags)
+{
+	int ret, fd;
+	struct elf_file *module;
+	const char *command;
+	struct module *mod = list_entry(list->next, struct module, list);
+	int rc = 0;
+
+	/* Take us off the list. */
+	list_del(&mod->list);
+
+	/* Do things we (or parent) depend on first. */
+	if (!list_empty(list)) {
+		modprobe_flags_t f = flags;
+		f &= ~mit_first_time;
+		f &= ~mit_ignore_commands;
+		if ((rc = insmod(list, NOFAIL(strdup("")), NULL,
+		       options, commands, "", warn, f)) != 0) {
+			error("Error inserting %s (%s): %s\n",
+				mod->modname, mod->filename,
+				insert_moderror(errno));
+			goto out_optstring;
+		}
+	}
+
+	fd = open_file(mod->filename);
+	if (fd < 0) {
+		error("Could not open '%s': %s\n",
+		      mod->filename, strerror(errno));
+		goto out_optstring;
+	}
+
+	/* Don't do ANYTHING if already in kernel. */
+	if (!(flags & mit_ignore_loaded)
+	    && module_in_kernel(newname ?: mod->modname, NULL) == 1) {
+		if (flags & mit_first_time)
+			error("Module %s already in kernel.\n",
+			      newname ?: mod->modname);
+		goto out_unlock;
+	}
+
+	command = find_command(mod->modname, commands);
+	if (command && !(flags & mit_ignore_commands)) {
+		close_file(fd);
+		do_command(mod->modname, command, flags & mit_dry_run, error,
+			   "install", cmdline_opts);
+		goto out_optstring;
+	}
+
+	module = grab_elf_file_fd(mod->filename, fd);
+	if (!module) {
+		error("Could not read '%s': %s\n", mod->filename,
+			(errno == ENOEXEC) ? "Invalid module format" :
+				strerror(errno));
+		goto out_unlock;
+	}
+	if (newname)
+		rename_module(module, mod->modname, newname);
+	if (flags & mit_strip_modversion)
+		module->ops->strip_section(module, "__versions");
+	if (flags & mit_strip_vermagic)
+		clear_magic(module);
+
+	/* Config file might have given more options */
+	optstring = add_extra_options(mod->modname, optstring, options);
+
+	info("insmod %s %s\n", mod->filename, optstring);
+
+	if (flags & mit_dry_run)
+		goto out;
+
+	ret = init_module(module->data, module->len, optstring);
+	if (ret != 0) {
+		if (errno == EEXIST) {
+			if (flags & mit_first_time)
+				error("Module %s already in kernel.\n",
+				      newname ?: mod->modname);
+			goto out_unlock;
+		}
+		/* don't warn noisely if we're loading multiple aliases. */
+		/* one of the aliases may try to use hardware we don't have. */
+		if ((error != warn) || (verbose))
+			error("Error inserting %s (%s): %s\n",
+			      mod->modname, mod->filename,
+			      insert_moderror(errno));
+		rc = 1;
+	}
+ out:
+	release_elf_file(module);
+ out_unlock:
+	close_file(fd);
+ out_optstring:
+	free(optstring);
+	return rc;
+}
+
+/* Do recursive removal. */
+static void rmmod(struct list_head *list,
+		  const char *name,
+		  struct module_command *commands,
+		  const char *cmdline_opts,
+		  errfn_t error,
+		  modprobe_flags_t flags)
+{
+	const char *command;
+	unsigned int usecount = 0;
+	struct module *mod = list_entry(list->next, struct module, list);
+
+	/* Take first one off the list. */
+	list_del(&mod->list);
+
+	if (!name)
+		name = mod->modname;
+
+	/* Even if renamed, find commands to orig. name. */
+	command = find_command(mod->modname, commands);
+	if (command && !(flags & mit_ignore_commands)) {
+		do_command(mod->modname, command, flags & mit_dry_run, error,
+			   "remove", cmdline_opts);
+		goto remove_rest;
+	}
+
+	if (module_in_kernel(name, &usecount) == 0)
+		goto nonexistent_module;
+
+	if (usecount != 0) {
+		if (!(flags & mit_ignore_loaded))
+			error("Module %s is in use.\n", name);
+		goto remove_rest;
+	}
+
+	info("rmmod %s\n", mod->filename);
+
+	if (flags & mit_dry_run)
+		goto remove_rest;
+
+	if (delete_module(name, O_EXCL) != 0) {
+		if (errno == ENOENT)
+			goto nonexistent_module;
+		error("Error removing %s (%s): %s\n",
+		      name, mod->filename,
+		      remove_moderror(errno));
+	}
+
+ remove_rest:
+	/* Now do things we depend. */
+	if (!list_empty(list)) {
+		flags &= ~mit_first_time;
+		flags &= ~mit_ignore_commands;
+		flags |= mit_ignore_loaded;
+
+		rmmod(list, NULL, commands, "", warn, flags);
+	}
+	return;
+
+nonexistent_module:
+	if (flags & mit_first_time)
+		fatal("Module %s is not in kernel.\n", mod->modname);
+	goto remove_rest;
+}
+
 static int handle_module(const char *modname,
 			  struct list_head *todo_list,
 			  const char *newname,
-			  int remove,
 			  char *options,
-			  int first_time,
-			  errfn_t error,
-			  int dry_run,
 			  struct module_options *modoptions,
 			  struct module_command *commands,
-			  int ignore_commands,
-			  int ignore_proc,
-			  int strip_vermagic,
-			  int strip_modversion,
 			  const char *cmdline_opts,
-			  int flags)
+			  errfn_t error,
+			  modprobe_flags_t flags)
 {
 	if (list_empty(todo_list)) {
 		const char *command;
@@ -1440,9 +1284,9 @@ static int handle_module(const char *modname,
 		/* The dependencies have to be real modules, but
 		   handle case where the first is completely bogus. */
 		command = find_command(modname, commands);
-		if (command && !ignore_commands) {
-			do_command(modname, command, dry_run, error,
-				   remove ? "remove":"install", cmdline_opts);
+		if (command && !(flags & mit_ignore_commands)) {
+			do_command(modname, command, flags & mit_dry_run, error,
+				   (flags & mit_remove) ? "remove":"install", cmdline_opts);
 			return 0;
 		}
 
@@ -1451,43 +1295,148 @@ static int handle_module(const char *modname,
 		return 1;
 	}
 
-	if (remove)
-		rmmod(todo_list, newname, first_time, error, dry_run,
-		      commands, ignore_commands, 0, cmdline_opts, flags);
-	else
+	if (flags & mit_remove) {
+		flags &= ~mit_ignore_loaded;
+		rmmod(todo_list, newname, commands, cmdline_opts, error, flags);
+	} else
 		insmod(todo_list, NOFAIL(strdup(options)), newname,
-		       first_time, error, dry_run, modoptions,
-		       commands, ignore_commands, ignore_proc, strip_vermagic,
-		       strip_modversion, cmdline_opts);
+		       modoptions, commands, cmdline_opts, error, flags);
 
 	return 0;
 }
 
-static struct option options[] = { { "verbose", 0, NULL, 'v' },
-				   { "version", 0, NULL, 'V' },
+int handle_builtin_module(const char *modname,
+                          errfn_t error,
+                          modprobe_flags_t flags)
+{
+	if (flags & mit_remove) {
+		error("Module %s is builtin\n", modname);
+		return 1;
+	} else if (flags & mit_first_time) {
+		error("Module %s already in kernel (builtin).\n", modname);
+		return 1;
+	} else if (flags & mit_ignore_loaded) {
+		/* --show-depends given */
+		info("builtin %s\n", modname);
+	}
+	return 0;
+}
+
+int do_modprobe(char *modname,
+		char *newname,
+		char *cmdline_opts,
+		const char *configname,
+		const char *dirname,
+		const char *aliasfilename,
+		const char *symfilename,
+		errfn_t error,
+		modprobe_flags_t flags)
+{
+	struct module_command *commands = NULL;
+	struct module_options *modoptions = NULL;
+	struct module_alias *aliases = NULL;
+	struct module_blacklist *blacklist = NULL;
+	LIST_HEAD(list);
+	int failed = 0;
+
+	/* Convert name we are looking for */
+	underscores(modname);
+
+	/* Returns the resolved alias, options */
+	parse_toplevel_config(configname, modname, 0,
+	     flags & mit_remove, &modoptions, &commands, &aliases, &blacklist);
+
+	/* Read module options from kernel command line */
+	parse_kcmdline(0, &modoptions);
+
+	/* No luck?  Try symbol names, if starts with symbol:. */
+	if (!aliases && strstarts(modname, "symbol:")) {
+		parse_config_file(symfilename, modname, 0,
+				  flags & mit_remove, &modoptions, &commands,
+				  &aliases, &blacklist);
+	}
+	if (!aliases) {
+		if(!strchr(modname, ':'))
+			read_depends(dirname, modname, &list);
+
+		/* We only use canned aliases as last resort. */
+		if (list_empty(&list)
+		    && !find_command(modname, commands))
+		{
+			read_aliases_file(aliasfilename,
+					  modname, 0, flags & mit_remove,
+					  &modoptions, &commands,
+					  &aliases, &blacklist);
+			/* builtin module? */
+			if (!aliases && module_builtin(dirname, modname) > 0) {
+				return handle_builtin_module(modname, error,
+						flags);
+			}
+		}
+	}
+
+	aliases = apply_blacklist(aliases, blacklist);
+	if(flags & mit_resolve_alias) {
+		for(; aliases; aliases=aliases->next)
+			printf("%s\n", aliases->module);
+		return 0;
+	}
+	if (aliases) {
+		errfn_t err = error;
+
+		/* More than one alias?  Don't bail out on failure. */
+		if (aliases->next)
+			err = warn;
+		while (aliases) {
+			/* Add the options for this alias. */
+			char *opts = NOFAIL(strdup(cmdline_opts));
+			opts = add_extra_options(modname,
+						 opts, modoptions);
+
+			read_depends(dirname, aliases->module, &list);
+			failed |= handle_module(aliases->module,
+				&list, newname, opts, modoptions,
+				commands, cmdline_opts, err, flags);
+
+			aliases = aliases->next;
+			INIT_LIST_HEAD(&list);
+		}
+	} else {
+		if (flags & mit_use_blacklist
+		    && find_blacklist(modname, blacklist))
+			return failed;
+
+		failed |= handle_module(modname, &list, newname, cmdline_opts,
+			modoptions, commands, cmdline_opts, error, flags);
+	}
+	return failed;
+}
+
+static struct option options[] = { { "version", 0, NULL, 'V' },
+				   { "verbose", 0, NULL, 'v' },
+				   { "quiet", 0, NULL, 'q' },
+				   { "syslog", 0, NULL, 's' },
+				   { "show", 0, NULL, 'n' },
+				   { "dry-run", 0, NULL, 'n' },
+				   { "show-depends", 0, NULL, 'D' },
+				   { "resolve-alias", 0, NULL, 'R' },
+				   { "dirname", 1, NULL, 'd' },
+				   { "set-version", 1, NULL, 'S' },
 				   { "config", 1, NULL, 'C' },
 				   { "name", 1, NULL, 'o' },
 				   { "remove", 0, NULL, 'r' },
-				   { "wait", 0, NULL, 'w' },
 				   { "showconfig", 0, NULL, 'c' },
-				   { "quiet", 0, NULL, 'q' },
-				   { "show", 0, NULL, 'n' },
-				   { "dry-run", 0, NULL, 'n' },
-				   { "syslog", 0, NULL, 's' },
-				   { "type", 1, NULL, 't' },
 				   { "list", 0, NULL, 'l' },
+				   { "type", 1, NULL, 't' },
 				   { "all", 0, NULL, 'a' },
 				   { "ignore-install", 0, NULL, 'i' },
 				   { "ignore-remove", 0, NULL, 'i' },
+				   { "use-blacklist", 0, NULL, 'b' },
 				   { "force", 0, NULL, 'f' },
 				   { "force-vermagic", 0, NULL, 1 },
 				   { "force-modversion", 0, NULL, 2 },
-				   { "set-version", 1, NULL, 'S' },
-				   { "show-depends", 0, NULL, 'D' },
-				   { "dirname", 1, NULL, 'd' },
 				   { "first-time", 0, NULL, 3 },
 				   { "dump-modversions", 0, NULL, 4 },
-				   { "use-blacklist", 0, NULL, 'b' },
 				   { NULL, 0, NULL, 0 } };
 
 int main(int argc, char *argv[])
@@ -1495,112 +1444,102 @@ int main(int argc, char *argv[])
 	struct utsname buf;
 	struct stat statbuf;
 	int opt;
-	int dump_only = 0;
-	int dry_run = 0;
-	int remove = 0;
+	int dump_config = 0;
 	int list_only = 0;
 	int all = 0;
-	int ignore_commands = 0;
-	int strip_vermagic = 0;
-	int strip_modversion = 0;
-	int ignore_proc = 0;
-	int first_time = 0;
 	int dump_modver = 0;
-	int use_blacklist = 0;
 	unsigned int i, num_modules;
 	char *type = NULL;
-	const char *config = NULL;
-	char *dirname = NULL;
-	char *optstring = NULL;
+	const char *configname = NULL;
+	char *basedir = "";
+	char *cmdline_opts = NULL;
 	char *newname = NULL;
-	char *aliasfilename, *symfilename;
+	char *dirname, *aliasfilename, *symfilename;
 	errfn_t error = fatal;
-	int flags = O_NONBLOCK|O_EXCL;
-	int was_error = 0;
+	int failed = 0;
+	modprobe_flags_t flags = 0;
 
 	/* Prepend options from environment. */
 	argv = merge_args(getenv("MODPROBE_OPTIONS"), argv, &argc);
 
 	uname(&buf);
-	while ((opt = getopt_long(argc, argv, "vVC:o:rnqQsclt:aifbwd:", options, NULL)) != -1){
+	while ((opt = getopt_long(argc, argv, "Vvqsnd:C:o:rclt:aibf", options, NULL)) != -1){
 		switch (opt) {
-		case 'v':
-			add_to_env_var("-v");
-			verbose = 1;
-			break;
 		case 'V':
 			puts(PACKAGE " version " VERSION);
 			exit(0);
-		case 'S':
-			strncpy(buf.release, optarg, sizeof(buf.release));
-			buf.release[sizeof(buf.release)-1] = '\0';
-			break;
-		case 'C':
-			config = optarg;
-			add_to_env_var("-C");
-			add_to_env_var(config);
+		case 'v':
+			add_to_env_var("-v");
+			verbose = 1;
 			break;
 		case 'q':
 			quiet = 1;
 			add_to_env_var("-q");
 			break;
+		case 's':
+			add_to_env_var("-s");
+			logging = 1;
+			break;
+		case 'n':
+			flags |= mit_dry_run;
+			break;
+		case 'd':
+			basedir = optarg;
+			break;
+		case 'S':
+			strncpy(buf.release, optarg, sizeof(buf.release));
+			buf.release[sizeof(buf.release)-1] = '\0';
+			break;
+		case 'C':
+			configname = optarg;
+			add_to_env_var("-C");
+			add_to_env_var(configname);
+			break;
 		case 'D':
-			dry_run = 1;
-			ignore_proc = 1;
+			flags |= mit_dry_run;
+			flags |= mit_ignore_loaded;
 			verbose = 1;
-			add_to_env_var("-D");
+			break;
+		case 'R':
+			flags |= mit_resolve_alias;
 			break;
 		case 'o':
 			newname = optarg;
 			break;
 		case 'r':
-			remove = 1;
+			flags |= mit_remove;
 			break;
 		case 'c':
-			dump_only = 1;
-			break;
-		case 't':
-			type = optarg;
+			dump_config = 1;
 			break;
 		case 'l':
 			list_only = 1;
 			break;
+		case 't':
+			type = optarg;
+			break;
 		case 'a':
 			all = 1;
 			error = warn;
 			break;
-		case 'n':
-			dry_run = 1;
-			break;
-		case 's':
-			add_to_env_var("-s");
-			logging = 1;
-			break;
 		case 'i':
-			ignore_commands = 1;
-			break;
-		case 'f':
-			strip_vermagic = 1;
-			strip_modversion = 1;
+			flags |= mit_ignore_commands;
 			break;
 		case 'b':
-			use_blacklist = 1;
+			flags |= mit_use_blacklist;
 			break;
-		case 'w':
-			flags &= ~O_NONBLOCK;
-			break;
-		case 'd':
-			nofail_asprintf(&dirname, "%s/%s/%s", optarg,
-					MODULE_DIR, buf.release);
+		case 'f':
+			flags |= mit_strip_vermagic;
+			flags |= mit_strip_modversion;
 			break;
 		case 1:
-			strip_vermagic = 1;
+			flags |= mit_strip_vermagic;
 			break;
 		case 2:
-			strip_modversion = 1;
+			flags |= mit_strip_modversion;
 			break;
 		case 3:
-			first_time = 1;
+			flags |= mit_first_time;
 			break;
 		case 4:
 			dump_modver = 1;
@@ -1616,11 +1555,10 @@ int main(int argc, char *argv[])
 		logging = 1;
 	}
 
-	if (argc < optind + 1 && !dump_only && !list_only && !remove)
+	if (argc < optind + 1 && !dump_config && !list_only)
 		print_usage(argv[0]);
 
-	if (!dirname)
-		nofail_asprintf(&dirname, "%s/%s", MODULE_DIR, buf.release);
+	nofail_asprintf(&dirname, "%s%s/%s", basedir, MODULE_DIR, buf.release);
 	nofail_asprintf(&aliasfilename, "%s/modules.alias", dirname);
 	nofail_asprintf(&symfilename, "%s/modules.symbols", dirname);
 
@@ -1634,13 +1572,13 @@ int main(int argc, char *argv[])
 	if (type)
 		fatal("-t only supported with -l");
 
-	if (dump_only) {
+	if (dump_config) {
 		struct module_command *commands = NULL;
 		struct module_options *modoptions = NULL;
 		struct module_alias *aliases = NULL;
 		struct module_blacklist *blacklist = NULL;
 
-		parse_toplevel_config(config, "", 1, 0, &modoptions,
+		parse_toplevel_config(configname, "", 1, 0, &modoptions,
 				      &commands, &aliases, &blacklist);
 		/* Read module options from kernel command line */
 		parse_kcmdline(1, &modoptions);
@@ -1651,99 +1589,25 @@ int main(int argc, char *argv[])
 		exit(0);
 	}
 
-	if (remove || all) {
+	if ((flags & mit_remove) || all) {
 		num_modules = argc - optind;
-		optstring = NOFAIL(strdup(""));
+		cmdline_opts = NOFAIL(strdup(""));
 	} else {
 		num_modules = 1;
-		optstring = gather_options(argv+optind+1);
+		cmdline_opts = gather_options(argv+optind+1);
 	}
 
 	/* num_modules is always 1 except for -r or -a. */
 	for (i = 0; i < num_modules; i++) {
-		struct module_command *commands = NULL;
-		struct module_options *modoptions = NULL;
-		struct module_alias *aliases = NULL;
-		struct module_blacklist *blacklist = NULL;
-		LIST_HEAD(list);
-		char *modulearg = argv[optind + i];
+		char *modname = argv[optind + i];
 
-		if (dump_modver) {
-			dump_modversions(modulearg, error);
-			continue;
-		}
+		if (dump_modver)
+			dump_modversions(modname, error);
+		else
+			failed |= do_modprobe(modname, newname, cmdline_opts,
+				configname, dirname, aliasfilename, symfilename,
+				error, flags);
 
-		/* Convert name we are looking for */
-		underscores(modulearg);
-
-		/* Returns the resolved alias, options */
-		parse_toplevel_config(config, modulearg, 0,
-		     remove, &modoptions, &commands, &aliases, &blacklist);
-
-		/* Read module options from kernel command line */
-		parse_kcmdline(0, &modoptions);
-
-		/* No luck?  Try symbol names, if starts with symbol:. */
-		if (!aliases && strstarts(modulearg, "symbol:")) {
-			parse_config_file(symfilename, modulearg, 0,
-					  remove, &modoptions, &commands,
-					  &aliases, &blacklist);
-		}
-		if (!aliases) {
-			if(!strchr(modulearg, ':'))
-				read_depends(dirname, modulearg, &list);
-
-			/* We only use canned aliases as last resort. */
-			if (list_empty(&list)
-			    && !find_command(modulearg, commands))
-			{
-				read_aliases_file(aliasfilename,
-						  modulearg, 0, remove,
-						  &modoptions, &commands,
-						  &aliases, &blacklist);
-			}
-		}
-
-		aliases = apply_blacklist(aliases, blacklist);
-		if (aliases) {
-			errfn_t err = error;
-
-			/* More than one alias?  Don't bail out on failure. */
-			if (aliases->next)
-				err = warn;
-			while (aliases) {
-				/* Add the options for this alias. */
-				char *opts = NOFAIL(strdup(optstring));
-				opts = add_extra_options(modulearg,
-							 opts, modoptions);
-
-				read_depends(dirname, aliases->module, &list);
-				if (handle_module(aliases->module, &list,
-					      newname, remove, opts,
-					      first_time, err,
-					      dry_run, modoptions,
-					      commands, ignore_commands,
-					      ignore_proc, strip_vermagic,
-					      strip_modversion,
-					      optstring, flags))
-					was_error = 1;
-
-				aliases = aliases->next;
-				INIT_LIST_HEAD(&list);
-			}
-		} else {
-			if (use_blacklist
-			    && find_blacklist(modulearg, blacklist))
-				continue;
-
-			if (handle_module(modulearg, &list, newname, remove,
-				      optstring, first_time, error, dry_run,
-				      modoptions, commands,
-				      ignore_commands, ignore_proc,
-				      strip_vermagic, strip_modversion,
-				      optstring, flags))
-				was_error = 1;
-		}
 	}
 	if (logging)
 		closelog();
@@ -1751,10 +1615,7 @@ int main(int argc, char *argv[])
 	free(dirname);
 	free(aliasfilename);
 	free(symfilename);
-	free(optstring);
+	free(cmdline_opts);
 
-	if (was_error)
-		exit(1);
-	else
-		exit(0);
+	exit(failed);
 }
